/*
 * Name:    mxbl_dcss.c
 *
 * Purpose: Handle DCSS events.
 *
 * Author:  William Lavender
 *
 *--------------------------------------------------------------------------
 *
 * Copyright 2008 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "mx_util.h"
#include "mx_record.h"
#include "mx_driver.h"
#include "mx_list.h"
#include "mx_callback.h"
#include "mx_bluice.h"

#include "mx_poison.h"

#include "mxbl.h"

/*------------------------------------------------------------------*/

typedef mx_status_type (MXBL_DCSS_MSG_HANDLER)( MXBL_EVENT * );

static MXBL_DCSS_MSG_HANDLER gtos_become_master;
static MXBL_DCSS_MSG_HANDLER gtos_configure_run;

static struct {
	char message_name[40];
	MXBL_DCSS_MSG_HANDLER *message_handler;
	mx_bool_type do_broadcast;
} dcss_handler_list[] = {
	{"gtos_abort_all", mxblh_abort_all, TRUE},
	{"gtos_become_master", gtos_become_master, FALSE},
	{"gtos_configure_run", gtos_configure_run, FALSE},
	{"gtos_read_ion_chambers", mxblh_read_ion_chambers, TRUE},
	{"gtos_set_shutter_state", mxblh_set_shutter_state, TRUE},
	{"gtos_start_motor_move", mxblh_start_motor_move, TRUE},
	{"gtos_start_operation", mxblh_start_operation, TRUE},
};

static int num_dcss_handlers = sizeof( dcss_handler_list )
				/ sizeof( dcss_handler_list[0] );

/*------------------------------------------------------------------*/

static mx_status_type
gtos_become_master( MXBL_EVENT *event )
{
	static const char fname[] = "gtos_become_master()";

	MXBL_EVENT_MANAGER *event_manager;
	MXBL_EVENT *old_master;
	MX_SOCKET *client_socket;
	int argc;
	char **argv;
	mx_status_type mx_status;

	event_manager = event->event_manager;
	client_socket = event->mx_socket;

	MX_DEBUG(-2,("%s invoked by Blu-Ice socket %d",
		fname, client_socket->socket_fd));

	mx_string_split( event->message_buffer, " ", &argc, &argv );

	if ( argc < 2 ) {
		mx_free(argv);

		return mx_error( MXE_NETWORK_IO_ERROR, fname,
		"A '%s' message was sent with less (%d) than the required "
		"number of arguments (2).", fname, argc );
	}

	/* Is there already another master? */

	if ( event_manager->master == NULL ) {
		/* If not, then become master immediately. */

		mx_free(argv);

		event_manager->master = event;

		mx_status = mxbl_send_message( client_socket,
					"stog_become_master",
					NULL, 0, MX_BLUICE_PROTOCOL_2 );
		return mx_status;
	}

	/* If we get here, there is already another master, so we check
	 * to see if we are 'force'-ing the issue.
	 */

	if ( strcmp( argv[1], "noforce" ) == 0 ) {
		/* Be nice and do not automatically take over. */

		mx_status = mxbl_send_message( client_socket,
					"stog_other_master",
					NULL, 0, MX_BLUICE_PROTOCOL_2 );
	} else
	if ( strcmp( argv[1], "force" ) == 0 ) {
		/* Become the master no matter what. */

		old_master = event_manager->master;

		event_manager->master = event;

		mx_status = mxbl_send_message( client_socket,
					"stog_become_master",
					NULL, 0, MX_BLUICE_PROTOCOL_2 );

		if ( mx_status.code != MXE_SUCCESS ) {
			mx_free(argv);
			return mx_status;
		}

		/* Tell the old master the bad news.
		 *
		 * Perhaps he can get the Doctor to help?
		 * After all, they went to school together
		 * on Gallifrey.
		 */

		mx_status = mxbl_send_message( old_master->mx_socket,
					"stog_become_slave",
					NULL, 0, MX_BLUICE_PROTOCOL_2 );
	} else {
		mx_status = mx_error( MXE_NETWORK_IO_ERROR, fname,
		"A '%s' message was received with invalid argument '%s'.",
			argv[0], argv[1] );

		mx_free(argv);
		return mx_status;
	}

	mx_free(argv);

	return mx_status;
}

/*------------------------------------------------------------------*/

#define FREE_CONFIGURE_RUN \
	do {					\
		mx_free(argv);			\
		mx_free(duplicate_message);	\
	} while (0)

static mx_status_type
gtos_configure_run( MXBL_EVENT *event )
{
	static const char fname[] = "gtos_configure_run()";

	MXBL_EVENT_MANAGER *event_manager;
	MX_SOCKET *client_socket;
	int argc;
	char **argv = NULL;
	char *duplicate_message = NULL;
	char *run_name, *run_status, *file_root, *directory, *scan_axis_name;
	unsigned long next_frame, run_label, start_frame_label;
	unsigned long num_energies, num_measurements, detector_mode;
	double start_position, end_position, step_size, wedge_size;
	double oscillation_time, detector_distance;
	double energy[5];
	int i, use_inverse_beam;
	char description[MXU_RECORD_DESCRIPTION_LENGTH+1];
	char buffer[100];
	MX_RECORD *ad_record, *energy_record, *distance_record;
	MX_RECORD *record_list, *current_record, *scan_record;
	mx_status_type mx_status;

	event_manager = event->event_manager;
	client_socket = event->mx_socket;
	record_list = event_manager->record_list;

	MX_DEBUG(-2,("%s invoked by Blu-Ice socket %d",
		fname, client_socket->socket_fd));

	/* Split up the message into its component parts. */

	duplicate_message = strdup( event->message_buffer );

	if ( duplicate_message == NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Ran out of memory trying to duplicate "
		"a Blu-Ice client message from socket %d.",
			client_socket->socket_fd );
	}
	
	mx_string_split( duplicate_message, " ", &argc, &argv );

	if ( argc < 23 ) {
		FREE_CONFIGURE_RUN;
		return mx_error( MXE_NETWORK_IO_ERROR, fname,
		"The gtos_configure_run command sent had only %d arguments.  "
		"The minimum number required is 23.  command = '%s'", 
			argc, event->message_buffer );
	}

	/* Find the area detector used by this run.  This will be the
	 * _last_ area detector record in the database.
	 */

	ad_record = NULL;

	current_record = record_list->previous_record;

	while( current_record != record_list ) {
		if ( (current_record->mx_superclass == MXR_DEVICE)
		  && (current_record->mx_class == MXC_AREA_DETECTOR) )
		{
			ad_record = current_record;
			break;
		}

		current_record = current_record->previous_record;
	}

	if ( ad_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NOT_FOUND, fname,
		"No area detector records were found in the MX database.");
	}

	MX_DEBUG(-2,
	("%s: FIXME: Do not hard code the names 'energy' and 'detector_z'.",
		fname));

	/* Find the energy record. */

	energy_record = mx_get_record( record_list, "energy" );

	if ( energy_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NOT_FOUND, fname,
		"No 'energy' motor was found in the MX database." );
	}

	/* Find the distance record. */

	distance_record = mx_get_record( record_list, "detector_z" );

	if ( distance_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NOT_FOUND, fname,
		"No 'distance' motor was found in the MX database." );
	}
	/* Start constructing the MX scan description. */

	run_name = argv[1];
	run_status = argv[2];
	next_frame = atol( argv[3] );
	run_label = atol( argv[4] );
	file_root = argv[5];
	directory = argv[6];
	start_frame_label = atol( argv[7] );
	scan_axis_name = argv[8];
	start_position = atof( argv[9] );
	end_position = atof( argv[10] );
	step_size = atof( argv[11] );
	wedge_size = atof( argv[12] );
	oscillation_time = atof( argv[13] );
	detector_distance = atof( argv[14] );
	num_energies = atol( argv[15] );
	energy[0] = atof( argv[16] );
	energy[1] = atof( argv[17] );
	energy[2] = atof( argv[18] );
	energy[3] = atof( argv[19] );
	energy[4] = atof( argv[20] );
	detector_mode = atol( argv[21] );
	use_inverse_beam = atoi( argv[22] );

	snprintf( description, sizeof(description),
	"%s scan ad_scan_class wedge_scan \"\" \"\" 1 1 1 %s 1 %s 0x0 0 "
	"preset_time \"%f timer1\" text %s/%s none \"\" %d %s %lu ",
		run_name, scan_axis_name, ad_record->name,
		oscillation_time, directory, file_root, use_inverse_beam,
		energy_record->name, num_energies );

	for ( i = 0; i < num_energies; i++ ) {
		snprintf( buffer, sizeof(buffer), "%f ", energy[i] );
		strlcat( description, buffer, sizeof(description) );
	}

	num_measurements = 1 +
	    mx_divide_safely( fabs(end_position - start_position), step_size );

	snprintf( buffer, sizeof(buffer), "1 %s %f %f %f %lu %f",
		distance_record->name, detector_distance,
		start_position, step_size, num_measurements, wedge_size );

	strlcat( description, buffer, sizeof(description) );

	MX_DEBUG(-2,("%s: scan description = '%s'", fname, description ));

	/* If an MX wedge scan with this name already exists, then delete it. */

	scan_record = mx_get_record( record_list, argv[1] );

	if ( scan_record != (MX_RECORD *) NULL ) {
		MX_DEBUG(-2,("%s: deleting old scan record '%s'",
			fname, argv[1] ));

		mx_status = mx_delete_record( scan_record );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;
	}

	/* Now create the new scan record. */

	mx_status = mx_create_record_from_description( record_list,
						description, &scan_record, 0 );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mx_status = mx_finish_record_initialization( scan_record );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Having replaced the MX scan, we finish by broadcasting
	 * the change to all clients.
	 */

	FREE_CONFIGURE_RUN;

	memcpy( event->message_buffer, "stog", 4 );

	mx_status = mxbl_broadcast_dcss_message( event_manager,
						event->message_buffer );
	return mx_status;
}

#undef FREE_CONFIGURE_RUN

/*------------------------------------------------------------------*/

void
mxbl_handle_dcss_messages( MXBL_EVENT *event )
{
	static const char fname[] = "mxbl_handle_dcss_messages()";

	MX_SOCKET *mx_socket;
	MXBL_DCSS_MSG_HANDLER *message_fn;
	double receive_timeout_in_seconds;
	double loop_timeout_in_seconds;
	MX_CLOCK_TICK loop_timeout_in_ticks;
	MX_CLOCK_TICK finish_tick, current_tick;
	int i, num_items, tick_comparison;
	long actual_data_length;
	char message_type_name[80];
	char message_type_format[20];
	mx_bool_type socket_is_open;
	mx_status_type mx_status;

	mx_socket = event->mx_socket;

	receive_timeout_in_seconds = 0.01;

	loop_timeout_in_seconds = 0.1;

	loop_timeout_in_ticks = mx_convert_seconds_to_clock_ticks(
						loop_timeout_in_seconds );

	snprintf( message_type_format, sizeof(message_type_format),
			"%%%lus", (unsigned long) sizeof(message_type_name) );

	MX_DEBUG(-2,("%s: DCSS socket %d is initialized.",
		fname, mx_socket->socket_fd ));

	current_tick = mx_current_clock_tick();

	finish_tick = mx_add_clock_ticks( current_tick, loop_timeout_in_ticks );

	for (;;) {
		MX_DEBUG(-2,("%s: Top of the for(;;) loop.", fname));

		/* See if we have timed out. */

		current_tick = mx_current_clock_tick();

		tick_comparison = mx_compare_clock_ticks( current_tick,
							finish_tick );

		if ( tick_comparison >= 0 ) {
			MX_DEBUG(-2,
			("%s: Exiting for(;;) loop after %g seconds.",
				fname, loop_timeout_in_seconds));

			break;	/* Exit the for(;;) loop. */
		}

		/* Give up if the socket is not open. */

		socket_is_open = mx_socket_is_open( mx_socket );

		if ( socket_is_open == FALSE ) {
			MX_DEBUG(-2,("%s: The DCSS socket is not open.",fname));

			mx_msleep(100);

			/* Go back to the top of the for(;;) loop. */

			continue;
		}

		/* Try to read a message for DCSS. */

		mx_status = mxbl_receive_message( mx_socket,
					event->message_buffer,
					sizeof(event->message_buffer),
					&actual_data_length,
					receive_timeout_in_seconds,
					MX_BLUICE_PROTOCOL_2 );

		if ( mx_status.code == MXE_NETWORK_CONNECTION_LOST ) {

			(void) mxbl_event_manager_unregister_socket_event(
								event );

			mx_socket_close( mx_socket );

			return;
		} else
		if ( mx_status.code != MXE_SUCCESS ) {
			mx_msleep(100);

			continue;
		}

		MX_DEBUG(-2,("%s: received '%s' from DCSS socket %d.",
			fname, event->message_buffer, mx_socket->socket_fd ));

		/* Figure out what kind of message this is. */

		num_items = sscanf( event->message_buffer,
				message_type_format,
				message_type_name );

		for ( i = 0; i < num_dcss_handlers; i++ ) {
			if ( strcmp( message_type_name,
				dcss_handler_list[i].message_name ) == 0 )
			{
				break;		/* Exit the for(i) loop. */
			}
		}

		if ( i >= num_dcss_handlers ) {
			(void) mx_error( MXE_NETWORK_IO_ERROR, fname,
			"Message type '%s' received in message '%s' from "
			"DCSS socket %d does not currently have a handler.",
				message_type_name, event->message_buffer,
				mx_socket->socket_fd );
		} else {
			/* Broadcast the message to all of the clients. */

			if ( dcss_handler_list[i].do_broadcast ) {
				(void) mxbl_broadcast_dcss_message(
						event->event_manager,
						event->message_buffer );
			}

			/* Invoke the DCSS handler. */

			message_fn = dcss_handler_list[i].message_handler;

			if ( message_fn == NULL ) {
				MX_DEBUG(-2,("%s: Message type '%s' SKIPPED.",
					fname, message_type_name));
			} else {
				MX_DEBUG(-2,("%s: Invoking handler for '%s'",
					fname, message_type_name));

				mx_status = (*message_fn)( event );
			}
		}
	}

	return;
}

/*------------------------------------------------------------------*/

static mx_status_type
mxblp_broadcast_handle_client( MX_LIST_ENTRY *list_entry,
				void *argument, void **not_used )
{
	static const char fname[] = "mxblp_broadcast_handle_client()";

	MXBL_EVENT *event;
	char *message_buffer;
	mx_status_type mx_status;

	event = list_entry->list_entry_data;

	/* Broadcast messages should only be sent to DCSS client sockets. */

	if ( event->socket_type != MXBLT_DCSS_CLIENT_SOCKET ) {
		return MX_SUCCESSFUL_RESULT;
	}

	message_buffer = argument;

#if 0
	MX_DEBUG(-2,("%s: event = %p, mx_socket = %p, socket_fd = %d",
		fname, event, event->mx_socket, event->mx_socket->socket_fd));
#endif

	MX_DEBUG(-2,("%s: sending '%s' to socket %d",
		fname, message_buffer, event->mx_socket->socket_fd));

	mx_status = mxbl_send_message( event->mx_socket, message_buffer,
					NULL, 0, MX_BLUICE_PROTOCOL_2 );

	/* Always send back a success status so that an error will not
	 * cause mx_list_traverse() to abort early.
	 */

	return MX_SUCCESSFUL_RESULT;
}

/*----*/

mx_status_type
mxbl_broadcast_dcss_message( MXBL_EVENT_MANAGER *event_manager,
				char *message_buffer )
{
	static const char fname[] = "mxbl_broadcast_dcss_message()";

	char *duplicate_message_buffer;
	mx_status_type mx_status;

	if ( event_manager == (MXBL_EVENT_MANAGER *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MXBL_EVENT_MANAGER pointer passed was NULL." );
	}
	if ( message_buffer == (char *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The message buffer pointer passed was NULL." );
	}

	duplicate_message_buffer = strdup( message_buffer );

	if ( duplicate_message_buffer == NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
			"Ran out of memory trying to duplicate a Blu-Ice "
			"message buffer." );
	}

	/* Change the first four letters of the message to 'stog'. */

	duplicate_message_buffer[0] = 's';
	duplicate_message_buffer[1] = 't';
	duplicate_message_buffer[2] = 'o';
	duplicate_message_buffer[3] = 'g';

	/* Send the modified message out to all of the DCSS clients. */

	MX_DEBUG(-2,("%s: broadcasting '%s' to all clients.",
		fname, duplicate_message_buffer));

	mx_status = mx_list_traverse( event_manager->event_list,
					mxblp_broadcast_handle_client,
					duplicate_message_buffer, NULL );

	mx_free( duplicate_message_buffer );

	return mx_status;
}

