/*
 * Name:    mxbl_dcss_server.c
 *
 * Purpose: Handle client connection to DCSS.
 *
 * Author:  William Lavender
 *
 *--------------------------------------------------------------------------
 *
 * Copyright 2008 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>

#include "mx_util.h"
#include "mx_driver.h"
#include "mx_list.h"
#include "mx_callback.h"
#include "mx_bluice.h"

#include "mx_motor.h"
#include "mx_relay.h"

#include "mx_scan.h"
#include "mx_scan_area_detector.h"
#include "sa_wedge.h"

#include "v_bluice_self_operation.h"
#include "v_bluice_string.h"

#include "mx_poison.h"

#include "mxbl.h"

/*------------------------------------------------------------------*/

static mx_bool_type
mxbl_dcss_timeout( const char *calling_fname,
		MX_SOCKET *mx_socket,
		double timeout_in_seconds,
		char *problem_source )
{
	unsigned long i, wait_ms, num_retries;
	long num_bytes_available;
	mx_status_type mx_status;

	wait_ms = 100;
	num_retries = (1000.0 * timeout_in_seconds) / (double) wait_ms;

	for ( i = 0; i < num_retries; i++ ) {
		mx_status = mx_socket_num_input_bytes_available(
					mx_socket, &num_bytes_available );

		if ( mx_status.code != MXE_SUCCESS )
			return TRUE;

		if ( num_bytes_available > 0 )
			break;		/* Exit the for() loop. */

		mx_msleep( wait_ms );
	}

	if ( i >= num_retries ) {
		(void) mx_error( MXE_TIMED_OUT, calling_fname,
		"Timed out waiting for %s from socket %d after %g seconds.",
		    problem_source, mx_socket->socket_fd, timeout_in_seconds );

		return TRUE;
	}

	return FALSE;
}

/*------------------------------------------------------------------*/

void
mxbl_handle_dcss_server_socket( MXBL_EVENT *dcss_server_event )
{
	static const char fname[] = "mxbl_handle_dcss_server_socket()";

	MX_SOCKET *dcss_server_socket, *client_socket;
	MXBL_EVENT *client_event;
	MXBL_EVENT_MANAGER *dcss_event_manager;
	char message[500];
	char buffer[50];
	char dhs_name[MXU_BLUICE_NAME_LENGTH+1];
	long actual_data_length;
	void *value_ptr;
	char *string_ptr;
	int saved_errno;
	char *error_string;
	mx_status_type mx_status;

	MX_RECORD *record_list, *current_record, *ion_chamber;
	MXBL_RECORD_INFO *rinfo;
	MXBL_TIMER_INFO *timer_info;
	MXBL_SCAN_INFO *scan_info;
	MX_MOTOR *motor;
	MX_RELAY *relay;

	MX_SCAN *scan;
	MX_AREA_DETECTOR_SCAN *ad_scan;
	MX_WEDGE_SCAN *wedge_scan;
	MX_RECORD *scan_axis_record;
	double end_position;
	unsigned long detector_mode, num_wedge_scans;

	MX_BLUICE_SELF_OPERATION *bluice_self_operation;
	MX_BLUICE_STRING *bluice_string;
	long i;

	struct sockaddr_in client_address;
	mx_socklen_t client_address_size = sizeof( struct sockaddr_in );

	dcss_server_socket = dcss_server_event->mx_socket;

	MX_DEBUG(-2,("%s invoked for socket %d.",
		fname, dcss_server_socket->socket_fd));

	/* Allocate memory for the DCSS client's MX_SOCKET structure. */

	client_socket = (MX_SOCKET *) malloc( sizeof(MX_SOCKET) );

	if ( client_socket == (MX_SOCKET *) NULL ) {
		(void) mx_error( MXE_OUT_OF_MEMORY, fname,
	    "Ran out of memory trying to allocate a Blu-Ice client socket." );

		return;
	}

	/* The server socket should be ready for an accept() operation,
	 * so let's try it.
	 */

	client_socket->socket_fd = accept( dcss_server_socket->socket_fd,
					(struct sockaddr *) &client_address,
					&client_address_size );

	saved_errno = mx_socket_check_error_status( &(client_socket->socket_fd),
					MXF_SOCKCHK_INVALID, &error_string );

	if ( saved_errno != 0 ) {
		mx_free( client_socket );

		(void) mx_error( MXE_NETWORK_IO_ERROR, fname,
		"An attempt to accept a connection on DCSS server port %d "
		"failed.  Errno = %d.  Error string = '%s'.",
			dcss_server_socket->socket_fd,
			saved_errno, error_string );

		return;
	}

	MX_DEBUG(-2,("%s: DCSS client socket %d created.",
		fname, client_socket->socket_fd));

	/* Ask the new client to send its client type. */

	mx_status = mxbl_send_message( client_socket,
					"stoc_send_client_type",
					NULL, 0, MX_BLUICE_PROTOCOL_2 );

	if ( mx_status.code != MXE_SUCCESS ) {
		mx_socket_close( client_socket );
		return;
	}

	/* The Blu-Ice client should respond with
	 * an htos_client_is_gui message.
	 */

	if ( mxbl_dcss_timeout( fname, client_socket,
				1.0, "htos_client_is_gui" ) )
	{
		return;
	}

	mx_status = mxbl_receive_message( client_socket,
					message, sizeof(message),
					&actual_data_length, 1.0,
					MX_BLUICE_PROTOCOL_1 );

	if ( mx_status.code != MXE_SUCCESS ) {
		(void) mx_error( mx_status.code, mx_status.location,
				mx_status.message );

		mx_socket_close( client_socket );

		return;
	}

	MX_DEBUG(-2,("%s: Client socket %d client type = '%s'",
		fname, client_socket->socket_fd, message));

	MX_DEBUG(-2,
		("%s: FIXME: We should do login authentication here.", fname));

	/* If the Blu-Ice client has been successfully authenticated,
	 * then we first create an event handler for the client.
	 */

	dcss_event_manager = dcss_server_event->event_manager;

	mx_status = mxbl_event_manager_register_socket_event(
					dcss_event_manager,
					client_socket,
					MXBLT_DCSS_CLIENT_SOCKET,
					mxbl_handle_dcss_messages,
					&client_event );

	if ( mx_status.code != MXE_SUCCESS ) {
		mx_socket_close( client_socket );

		return;
	}

	/* If the event handler was successfully created, then
	 * assign the client a client number.
	 */

	if ( dcss_server_event->client_number == 0 ) {
		dcss_server_event->client_number = 1;
	}

	client_event->client_number = dcss_server_event->client_number;

	dcss_server_event->client_number++;

	/* Tell the client that it has successfully logged in. */

	snprintf( message, sizeof(message),
		"stog_login_complete %lu", client_event->client_number );

	mx_status = mxbl_send_message( client_socket, message,
					NULL, 0, MX_BLUICE_PROTOCOL_2 );

	if ( mx_status.code != MXE_SUCCESS ) {
		mx_socket_close( client_socket );

		return;
	}

	/* Get the DHS name from the DCSS event manager structure. */

	strlcpy( dhs_name, dcss_event_manager->dhs_name, sizeof(dhs_name) );

	/* Tell the client that the server 'dhs_name' is online. */

	snprintf( message, sizeof(message),
		"stog_configure_hardware_host %s localhost online", dhs_name );

	mx_status = mxbl_send_message( client_socket, message,
					NULL, 0, MX_BLUICE_PROTOCOL_2 );

	if ( mx_status.code != MXE_SUCCESS ) {
		mx_socket_close( client_socket );

		return;
	}

	/* Tell the client how many wedge scans there are in the
	 * database.  Blu-Ice considers the scan called 'run0'
	 * to be special.  'run0' must always be present, but it
	 * is not counted in the number of runs to be sent to the
	 * client.
	 */

	record_list = dcss_event_manager->record_list;

	current_record = record_list->next_record;

	num_wedge_scans = 0;

	while( current_record != record_list ) {
		if ( current_record->mx_type == MXS_AD_WEDGE ) {
			if ( strcmp(current_record->name, "run0") != 0 ) {
				num_wedge_scans++;
			}
		}
		current_record = current_record->next_record;
	}

	snprintf( message, sizeof(message),
		"stog_configure_runs %lu 0 0 0", num_wedge_scans );

	mx_status = mxbl_send_message( client_socket, message, NULL, 0,
						MX_BLUICE_PROTOCOL_2 );

	if ( mx_status.code != MXE_SUCCESS ) {
		mx_socket_close( client_socket );

		return;
	}

	/* We finish by sending the status of the beamline to the client now. */

	current_record = record_list->next_record;

	while( current_record != record_list ) {
	    /* We are only interested in records that have a non-NULL
	     * MXBL_RECORD_INFO pointer.
	     */

	    rinfo = current_record->application_ptr;

	    if ( rinfo != (MXBL_RECORD_INFO *) NULL ) {
		switch( current_record->mx_superclass ) {
		case MXR_DEVICE:
		    switch( current_record->mx_class ) {
		    case MXC_MOTOR:
		    	mx_status = mx_motor_get_extended_status(
						current_record, NULL, NULL );

			if ( mx_status.code == MXE_SUCCESS ) {
			    motor = current_record->record_class_struct;

		    	    snprintf( message, sizeof(message),
			    "stog_configure_real_motor %s %s %s "
					"%f %f %f %f %f %f %f 1 1 0 0 0 0",
				current_record->name,
				dhs_name,
				current_record->name,
				motor->position,
				motor->positive_limit,
				motor->negative_limit,
				motor->scale,
				motor->speed,
				motor->acceleration_time,
				motor->backlash_correction );

			    mx_status = mxbl_send_message( client_socket,
			    		message, NULL, 0, MX_BLUICE_PROTOCOL_2);
			}
		    	break;
		    case MXC_RELAY:
		    	mx_status = mx_get_relay_status( current_record, NULL );

			if ( mx_status.code == MXE_SUCCESS ) {
			    relay = current_record->record_class_struct;

			    switch( relay->relay_status ) {
			    case MXF_RELAY_IS_CLOSED:
			    	snprintf( message, sizeof(message),
				"stog_configure_shutter %s %s open",
					current_record->name, dhs_name );
			    default:
			    	snprintf( message, sizeof(message),
				"stog_configure_shutter %s %s open",
					current_record->name, dhs_name );
			    }

			    mx_status = mxbl_send_message( client_socket,
			    		message, NULL, 0, MX_BLUICE_PROTOCOL_2);
			}
		    	break;
		    case MXC_TIMER:
		    	mx_status = mxbl_find_ion_chambers( current_record );

			if ( mx_status.code == MXE_SUCCESS ) {

			    timer_info = rinfo->rinfo_private;

			    if ( timer_info == (MXBL_TIMER_INFO *) NULL ) {
			    	mx_error( MXE_CORRUPT_DATA_STRUCTURE,
					fname, "The MXBL_TIMER_INFO pointer "
					"for timer record '%s' is NULL.",
						current_record->name );
				return;
			    }

		    	    MX_DEBUG(-2,
			    ("%s: timer '%s', num_ion_chambers = %lu",
				fname, current_record->name,
				timer_info->num_ion_chambers));

			    for ( i = 0; i < timer_info->num_ion_chambers; i++ )
			    {
			        ion_chamber = timer_info->ion_chamber_array[i];

				snprintf( message, sizeof(message),
			     "stog_configure_ion_chamber %s %s %s %ld %s count",
					ion_chamber->name,
					dhs_name,
					"hex1", i,
					"rtc1" );

			        mx_status = mxbl_send_message( client_socket,
			    		message, NULL, 0, MX_BLUICE_PROTOCOL_2);
			    }
			}
		    	break;
		    }
		    break;

		case MXR_VARIABLE:
		    switch( current_record->mx_type ) {
		    case MXV_BLUICE_SELF_OPERATION:
		    	bluice_self_operation =
				current_record->record_type_struct;

		        snprintf( message, sizeof(message),
			"stog_configure_operation %s %s",
				bluice_self_operation->bluice_name,
				bluice_self_operation->bluice_server_name );

		        mx_status = mxbl_send_message( client_socket,
			    		message, NULL, 0, MX_BLUICE_PROTOCOL_2);
		        break;
		    case MXV_BLUICE_SELF_STRING:
		        mx_status = mx_get_variable_pointer( current_record,
								&value_ptr );

			if ( mx_status.code == MXE_SUCCESS ) {
			    bluice_string = current_record->record_type_struct;

			    string_ptr = (char *) value_ptr;

			    if ( string_ptr[0] == '\0' ) {
			        snprintf( message, sizeof(message),
				"stog_configure_string %s self {}",
				    bluice_string->bluice_name );
			    } else {
				snprintf( message, sizeof(message),
				"stog_configure_string %s self %s",
				    bluice_string->bluice_name, string_ptr );
			    }

		            mx_status = mxbl_send_message( client_socket,
			    		message, NULL, 0, MX_BLUICE_PROTOCOL_2);
			}
			break;
		    }
		    break;

		case MXR_SCAN:
		    switch( current_record->mx_type ) {
		    case MXS_AD_WEDGE:
		    	scan = current_record->record_superclass_struct;

			ad_scan = current_record->record_class_struct;

			wedge_scan = current_record->record_type_struct;

			scan_info = rinfo->rinfo_private;

			if ( scan_info == (MXBL_SCAN_INFO *) NULL ) {
				mx_error( MXE_INITIALIZATION_ERROR, fname,
				"The MXBL_SCAN_INFO pointer for record '%s' "
				"is NULL.", current_record->name );
				break;
			}

			/* Find the scan motor. */

			if ( scan->num_motors == 0 ) {
				mx_error( MXE_UNSUPPORTED, fname,
				"Scan '%s' is configured for 0 motors "
				"which is not supported for Blu-Ice "
				"area detector runs.",
					current_record->name );
				break;
			}

			scan_axis_record = scan->motor_record_array[0];

			/* Compute the end position for the scan. */

			end_position = ad_scan->start_position[0]
				+ ad_scan->step_size[0]
					* ( ad_scan->num_frames[0] - 1 );

			/* Verify the presence of the detector motor. */

			if ( ad_scan->num_static_motors == 0 ) {
				mx_error( MXE_UNSUPPORTED, fname,
				"Scan '%s' does not have a detector distance "
				"record configured.", current_record->name );
			}

			/* Construct the configure message. */

		    	snprintf( message, sizeof(message),
			"stog_configure_run %s %s %lu %lu %s %s %lu %s "
			"%f %f %f %f %f %f %ld ",
				current_record->name,
				scan_info->run_status,
				scan_info->next_frame,
				scan_info->run_label,
				scan_info->file_root,
				scan_info->directory,
				scan_info->start_frame_label,
				scan_axis_record->name,
				ad_scan->start_position[0],
				end_position,
				ad_scan->step_size[0],
				wedge_scan->wedge_size,
				mx_scan_get_measurement_time( scan ),
				ad_scan->static_motor_positions[0],
				ad_scan->num_energies );

			for ( i = 0; i < ad_scan->num_energies; i++ ) {
				snprintf( buffer, sizeof(buffer), "%f ", 
					ad_scan->energy_array[i] );

				strlcat( message, buffer, sizeof(message) );
			}

			for ( i = ad_scan->num_energies; i < 5; i++ ) {
				strlcat( message, "0 ", sizeof(message) );
			}

			detector_mode = 2;  /* FIXME: Don't use fixed value. */

			snprintf( buffer, sizeof(buffer), "%lu %d",
				detector_mode, ad_scan->use_inverse_beam );

			strlcat( message, buffer, sizeof(message) );

			/* Send the configure message. */

		        mx_status = mxbl_send_message( client_socket,
			    		message, NULL, 0, MX_BLUICE_PROTOCOL_2);
		    	break;
		    }
		    break;
		}
	    }

	    current_record = current_record->next_record;
	}

	return;
}

/*------------------------------------------------------------------*/

mx_status_type
mxbl_register_dcss_callbacks( MXBL_EVENT_MANAGER *event_manager,
				MX_RECORD *record_list )
{
	static const char fname[] = "mxbl_register_dcss_callbacks()";

	MX_RECORD *current_record;
	mx_status_type mx_status;

	MX_DEBUG(-2,("%s invoked.", fname));

	/* Make sure this is the list head. */

	record_list = record_list->list_head;

	/* Walk through all of the records setting up DCSS-related callbacks
	 * as we go.
	 */

	/* FIXME: We should provide a way to only selectively add callbacks. */

	MX_DEBUG(-2,("%s: FIXME: Selective callbacks needed.", fname));

	current_record = record_list->next_record;

	mx_status = MX_SUCCESSFUL_RESULT;

	while ( current_record != record_list ) {

		MX_DEBUG(-2,("%s: current_record = '%s'",
				fname, current_record->name));

		mx_status = mxbl_setup_record_info( event_manager,
						current_record,
						current_record->name );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		switch( current_record->mx_superclass ) {
		case MXR_DEVICE:
			switch( current_record->mx_class ) {
			case MXC_MOTOR:
				mx_status = mxbl_register_field_callback(
					current_record, "extended_status" );
				break;
			case MXC_RELAY:
				mx_status = mxbl_register_field_callback(
					current_record, "relay_status" );
				break;
			case MXC_TIMER:
				mx_status = mxbl_register_field_callback(
					current_record, "busy" );
				break;
			}
			break;

		case MXR_VARIABLE:


		default:
			break;
		}

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		current_record = current_record->next_record;
	}

	return MX_SUCCESSFUL_RESULT;
}

