/*
 * Name:    mxbl_dhs.c
 *
 * Purpose: Handle DHS events.
 *
 * Author:  William Lavender
 *
 *--------------------------------------------------------------------------
 *
 * Copyright 2008 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>

#include "mx_util.h"
#include "mx_record.h"
#include "mx_driver.h"
#include "mx_list.h"
#include "mx_socket.h"
#include "mx_callback.h"
#include "mx_process.h"

#include "mx_scaler.h"
#include "mx_bluice.h"
#include "v_bluice_string.h"
#include "v_bluice_self_operation.h"

#include "mx_poison.h"

#include "mxbl.h"

/*------------------------------------------------------------------*/

typedef mx_status_type (MXBL_DHS_MSG_HANDLER)( MXBL_EVENT * );

static MXBL_DHS_MSG_HANDLER stoh_configure_real_motor;
static MXBL_DHS_MSG_HANDLER stoh_read_ion_chambers;
static MXBL_DHS_MSG_HANDLER stoh_register_ion_chamber;
static MXBL_DHS_MSG_HANDLER stoh_register_operation;
static MXBL_DHS_MSG_HANDLER stoh_register_real_motor;
static MXBL_DHS_MSG_HANDLER stoh_register_shutter;
static MXBL_DHS_MSG_HANDLER stoh_register_string;
static MXBL_DHS_MSG_HANDLER stoh_set_motor_position;

static struct {
	char message_name[40];
	MXBL_DHS_MSG_HANDLER *message_handler;
} dhs_handler_list[] = {
	{"stoh_abort_all", mxblh_abort_all},
	{"stoh_configure_real_motor", stoh_configure_real_motor},
	{"stoh_read_ion_chambers", stoh_read_ion_chambers},
	{"stoh_register_ion_chamber", stoh_register_ion_chamber},
	{"stoh_register_operation", stoh_register_operation},
	{"stoh_register_real_motor", stoh_register_real_motor},
	{"stoh_register_shutter", stoh_register_shutter},
	{"stoh_register_string", stoh_register_string},
	{"stoh_set_motor_position", stoh_set_motor_position},
	{"stoh_set_shutter_state", mxblh_set_shutter_state},
	{"stoh_start_motor_move", mxblh_start_motor_move},
	{"stoh_start_operation", mxblh_start_operation},
};

static int num_dhs_handlers = sizeof( dhs_handler_list )
				/ sizeof( dhs_handler_list[0] );

/*------------------------------------------------------------------*/

static void
mxbl_handle_dhs_message_initialization( MXBL_EVENT *event )
{
	static const char fname[] = "mxbl_handle_dhs_message_initialization()";

	MX_SOCKET *mx_socket;
	char message[500];
	long actual_message_length;
	mx_status_type mx_status;

	mx_socket = event->mx_socket;

	MX_DEBUG(-2,("%s: mx_socket->socket_fd = %d",
		fname, mx_socket->socket_fd));

	/* Handshaking with the DCSS should start with DCSS sending us
	 * an stoc_send_client_type message using Blu-Ice protocol 1.
	 */

	mx_status = mxbl_receive_message( mx_socket,
					message, sizeof(message),
					&actual_message_length, 1.0,
					MX_BLUICE_PROTOCOL_1 );

	if ( mx_status.code != MXE_SUCCESS )
		return;

	MX_DEBUG(-2,("%s: actual_message_length = %ld",
		fname, actual_message_length));

	MX_DEBUG(-2,("%s: received '%s' from Blu-Ice socket %d",
		fname, message, mx_socket->socket_fd));

	if ( strcmp(message, "stoc_send_client_type") != 0 ) {
		(void) mx_error( MXE_NETWORK_IO_ERROR, fname,
		"The first message sent by DCSS socket %d was _not_ "
		"an stoc_send_client_type message.  Instead, "
		"the message was '%s'.",
			mx_socket->socket_fd, message );
	}

	/* Send back a message stating that we are using DCS protocol 2.0. */

	snprintf( message, sizeof(message),
		"htos_client_type_is_hardware %s dcsProtocol_2.0",
		event->event_manager->dhs_name );

	MX_DEBUG(-2,("%s: sending '%s' to Blu-Ice socket %d",
		fname, message, mx_socket->socket_fd ));

	mx_status = mxbl_send_message( mx_socket, message,
					NULL, 0, MX_BLUICE_PROTOCOL_1 );

	if ( mx_status.code != MXE_SUCCESS )
		return;

	/* If the transmission of this message went well, then mark
	 * the DHS socket as having been initialized.
	 */

	event->protocol_initialized = TRUE;

	return;
}

/*------------------------------------------------------------------*/

static mx_status_type
stoh_configure_real_motor( MXBL_EVENT *event )
{
	static const char fname[] = "stoh_configure_real_motor()";

	MX_SOCKET *mx_socket;
	char *ptr, *token_ptr;
	char *dcss_motor_name, *dhs_name, *mx_motor_name;
	MX_RECORD *motor_record;
	MXBL_RECORD_INFO *rinfo;

	mx_socket = event->mx_socket;

	/* Skip over the command name. */

	ptr = event->message_buffer;

	token_ptr = mx_string_token( &ptr, " " );

	if ( token_ptr == NULL ) {
		return mx_error( MXE_UNPARSEABLE_STRING, fname,
		"The DHS message buffer received for socket %d was empty.  "
		"This should not be able to happen.", mx_socket->socket_fd );
	}

	/* The next token should be the name of the motor in the DCSS. */

	dcss_motor_name = mx_string_token( &ptr, " " );

	if ( dcss_motor_name == NULL ) {
		return mx_error( MXE_UNPARSEABLE_STRING, fname,
		"No DCSS motor name was seen in the stoh_configure_real_motor "
		"command received by DHS socket %d.",
			mx_socket->socket_fd );
	}

	/* The next token should be the DHS name. */

	dhs_name = mx_string_token( &ptr, " " );

	if ( dhs_name == NULL ) {
		return mx_error( MXE_UNPARSEABLE_STRING, fname,
		"No Blu-Ice DHS name was seen in the stoh_configure_real_motor "
		"command received by DHS socket %d.",
			mx_socket->socket_fd );
	}

	/* FIXME: We should check the DHS name to verify that DCSS
	 * is referring to us or not.
	 */

	/* The next token should be the local MX motor name. */

	mx_motor_name = mx_string_token( &ptr, " " );

	if ( mx_motor_name == NULL ) {
		return mx_error( MXE_UNPARSEABLE_STRING, fname,
		"No MX motor name was seen in the stoh_configure_real_motor "
		"command received by DHS socket %d.",
			mx_socket->socket_fd );
	}

	MX_DEBUG(-2,
	("%s: dcss_motor_name = '%s', dhs_name = '%s', mx_motor_name = '%s'",
		fname, dcss_motor_name, dhs_name, mx_motor_name));

	/* Find the motor in the MX database. */

	motor_record = mx_get_record( event->event_manager->record_list,
						mx_motor_name );

	if ( motor_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NOT_FOUND, fname,
		"MX motor '%s' requested by DHS socket %d does not exist.",
			mx_motor_name, mx_socket->socket_fd );
	}

	/* FIXME: For now we ignore the configuration change and 
	 * just display the configuration string sent.
	 */

	MX_DEBUG(-2,("%s: The configuration change requested by DHS socket %d "
	"for MX motor '%s' was '%s'",
		fname, mx_socket->socket_fd, motor_record->name, ptr ));

	/* Arrange for the position callback to update the current motor
	 * position by setting the operation_in_progress flag.
	 */

	rinfo = motor_record->application_ptr;

	if ( rinfo == (MXBL_RECORD_INFO *) NULL ) {
		return mx_error( MXE_INITIALIZATION_ERROR, fname,
		"The MXBL_RECORD_INFO structure for motor '%s' has not "
		"been allocated.", motor_record->name );
	}

	rinfo->operation_in_progress = TRUE;

	return MX_SUCCESSFUL_RESULT;
}

/*------------------------------------------------------------------*/

static void
mxblp_free_ion_chambers( MX_RECORD **ion_chamber_array,
				MX_RECORD *timer_record )
{
	MXBL_RECORD_INFO *rinfo;
	MXBL_TIMER_INFO *timer_info;

	mx_free( ion_chamber_array );

	if ( timer_record != NULL ) {
		rinfo = timer_record->application_ptr;

		if ( rinfo != NULL ) {
			timer_info = rinfo->rinfo_private;

			if ( timer_info != NULL ) {
				timer_info->num_ion_chambers = 0;

				mx_free( timer_info->ion_chamber_array );
			}
		}
	}
}

static mx_status_type
stoh_read_ion_chambers( MXBL_EVENT *event )
{
	static const char fname[] = "stoh_read_ion_chambers()";

	MX_SOCKET *mx_socket;
	MX_RECORD *record_list, *ion_chamber_record, *timer_record;
	MX_ANALOG_INPUT *ainput;
	MX_SCALER *scaler;
	MXBL_RECORD_INFO *rinfo;
	MXBL_TIMER_INFO *timer_info;
	char *endptr, *ptr, *token_ptr, *time_string, *repeat_count_string;
	char *ion_chamber_name;
	long num_ion_chambers, num_blocks, block_size, array_size;
	unsigned long repeat_count;
	MX_RECORD **ion_chamber_array, **new_ion_chamber_array;
	double measurement_time;
	mx_status_type mx_status;

	MX_DEBUG(-2,("%s invoked.", fname));

	mx_socket = event->mx_socket;

	record_list = event->event_manager->record_list;

	/* Skip over the command name. */

	ptr = event->message_buffer;

	MX_DEBUG(-2,("%s: message_buffer = '%s'", fname, ptr));

	token_ptr = mx_string_token( &ptr, " " );

	if ( token_ptr == NULL ) {
		return mx_error( MXE_UNPARSEABLE_STRING, fname,
		"The DHS message buffer received for socket %d was empty.  "
		"This should not be able to happen.", mx_socket->socket_fd );
	}

	/* The next token should be the measurement time. */

	time_string = mx_string_token( &ptr, " " );

	if ( time_string == NULL ) {
		return mx_error( MXE_UNPARSEABLE_STRING, fname,
		"No abort type was seen in the stoh_read_ion_chambers "
		"command received by DHS socket %d.",
			mx_socket->socket_fd );
	}

	/* The next token is the repeat count string. */

	repeat_count_string = mx_string_token( &ptr, " " );

	if ( repeat_count_string == NULL ) {
		return mx_error( MXE_UNPARSEABLE_STRING, fname,
		"No repeat count was seen in the stoh_read_ion_chambers "
		"command received by DHS socket %d.",
			mx_socket->socket_fd );
	}

	repeat_count = mx_string_to_unsigned_long( repeat_count_string );

	if ( repeat_count != 0 ) {
		mx_warning( "The MX Blu-Ice server does not currently "
		"support repeat counts for the stoh_read_ion_chambers "
		"command." );
	}

	/* Allocate an initial array of ion chamber records. */

	num_blocks = 1;
	block_size = 5;

	array_size = num_blocks * block_size;

	ion_chamber_array = (MX_RECORD **)
		malloc( array_size * sizeof(MX_RECORD *) );

	if ( ion_chamber_array == (MX_RECORD **) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Ran out of memory trying to allocate a %ld-element "
		"array of ion chamber records.", array_size );
	}

	timer_record = NULL;

	num_ion_chambers = 0;

	/* Loop through the list of ion chamber records. */

	for (;;) {
		ion_chamber_name = mx_string_token( &ptr, " " );

		if ( ion_chamber_name == NULL ) {
			/* We have now run out of ion chambers,
			 * so break out of the for(;;) loop.
			 */

			break;
		}

		if ( num_ion_chambers >= array_size ) {
			num_blocks++;

			array_size = num_blocks * block_size;

			MX_DEBUG(-2,("%s: realloc()-ing to %ld",
				fname, array_size));

			new_ion_chamber_array = realloc( ion_chamber_array,
								array_size );

			if ( new_ion_chamber_array == NULL ) {
				mxblp_free_ion_chambers( ion_chamber_array,
							timer_record );

				return mx_error( MXE_OUT_OF_MEMORY, fname,
				"Ran out of memory trying to increase the "
				"number of records in the ion chamber array "
				"to %ld.", array_size );
			}

			ion_chamber_array = new_ion_chamber_array;
		}

		/* Find the ion chamber in the MX database. */

		ion_chamber_record = mx_get_record( record_list,
						ion_chamber_name );

		if ( ion_chamber_record == (MX_RECORD *) NULL ) {
			mxblp_free_ion_chambers( ion_chamber_array,
							timer_record );

			return mx_error( MXE_NOT_FOUND, fname,
			"Ion chamber '%s' requested by DHS socket %d "
			"is not found in the MX database.",
				ion_chamber_name, mx_socket->socket_fd );
		}

		switch( ion_chamber_record->mx_class ) {
		case MXC_ANALOG_INPUT:
			ainput = ion_chamber_record->record_class_struct;

			if ( ainput->timer_record == NULL ) {
				mxblp_free_ion_chambers( ion_chamber_array,
							timer_record );

				return mx_error(MXE_INITIALIZATION_ERROR, fname,
				"MX analog input '%s' does not have an "
				"associated timer record.",
					ion_chamber_record->name );
			}

			if ( timer_record == NULL ) {
				timer_record = ainput->timer_record;
			} else
			if ( timer_record != ainput->timer_record ) {
				mxblp_free_ion_chambers( ion_chamber_array,
							timer_record );

				return mx_error(MXE_INITIALIZATION_ERROR, fname,
				"MX ion chamber '%s' uses a timer '%s' "
				"that is not the same as the timer '%s' used "
				"by the first ion chamber '%s'.",
					ion_chamber_record->name,
					ainput->timer_record->name,
					timer_record->name,
					ion_chamber_array[0]->name );
			}

			ion_chamber_array[num_ion_chambers]
				= ion_chamber_record;

			num_ion_chambers++;
			break;

		case MXC_SCALER:
			mx_status = mx_scaler_clear( ion_chamber_record );

			if ( mx_status.code != MXE_SUCCESS ) {
				mxblp_free_ion_chambers( ion_chamber_array,
							timer_record );

				return mx_error( mx_status.code, fname,
				"Unable to clear scaler '%s'.",
					ion_chamber_record->name );
			}

			scaler = ion_chamber_record->record_class_struct;

			if ( scaler->timer_record == NULL ) {
				mxblp_free_ion_chambers( ion_chamber_array,
							timer_record );

				return mx_error(MXE_INITIALIZATION_ERROR, fname,
				"MX scaler '%s' does not have an "
				"associated timer record.",
					ion_chamber_record->name );
			}

			if ( timer_record == NULL ) {
				timer_record = scaler->timer_record;
			} else
			if ( timer_record != scaler->timer_record ) {
				mxblp_free_ion_chambers( ion_chamber_array,
							timer_record );

				return mx_error(MXE_INITIALIZATION_ERROR, fname,
				"MX ion chamber '%s' uses a timer '%s' "
				"that is not the same as the timer '%s' used "
				"by the first ion chamber '%s'.",
					ion_chamber_record->name,
					scaler->timer_record->name,
					timer_record->name,
					ion_chamber_array[0]->name );
			}

			ion_chamber_array[num_ion_chambers]
				= ion_chamber_record;

			num_ion_chambers++;
			break;

		default:
			mxblp_free_ion_chambers( ion_chamber_array,
						timer_record );

			return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"MX record '%s' of type '%s' cannot be used as "
			"a Blu-Ice ion chamber.",
				ion_chamber_record->name,
				mx_get_driver_name( ion_chamber_record ) );
			break;
		}
	}

	if ( num_ion_chambers == 0 ) {
		mxblp_free_ion_chambers( ion_chamber_array, timer_record );

		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"No ion chambers were specified by the "
		"stoh_read_ion_chambers command." );
	}

#if 1
	{
		long i;

		MX_DEBUG(-2,("%s: num_ion_chambers = %ld",
			fname, num_ion_chambers));

		for ( i = 0; i < num_ion_chambers; i++ ) {
			MX_DEBUG(-2,("%s:   ion_chamber[%ld] = '%s'",
				fname, i, ion_chamber_array[i]->name ));
		}
	}
#endif
	/* Get the measurement time. */

	measurement_time = strtod( time_string, &endptr );

	if ( *endptr != '\0' ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"The measurement time '%s' requested by DHS socket %d "
		"for ion chamber '%s' is not a legal floating point number.",
			time_string, mx_socket->socket_fd,
			ion_chamber_array[0]->name );
	}

	/* Attach the array of ion chambers to the MXBL_RECORD_INFO structure
	 * of the timer record.
	 */

	rinfo = timer_record->application_ptr;

	if ( rinfo == (MXBL_RECORD_INFO *) NULL ) {
		mxblp_free_ion_chambers( ion_chamber_array, timer_record );

		return mx_error( MXE_INITIALIZATION_ERROR, fname,
		"MX timer '%s' does not have an associated "
		"MXBL_RECORD_INFO structure.  This should not happen.",
			timer_record->name );
	}

	timer_info = rinfo->rinfo_private;

	if ( timer_info == (MXBL_TIMER_INFO *) NULL ) {
		mxblp_free_ion_chambers( ion_chamber_array, timer_record );

		return mx_error( MXE_INITIALIZATION_ERROR, fname,
		"MX timer '%s' does not have an associated "
		"MXBL_TIMER_INFO structure.  This should not happen.",
			timer_record->name );
	}

	mx_free( timer_info->ion_chamber_array );

	timer_info->num_ion_chambers  = num_ion_chambers;
	timer_info->ion_chamber_array = ion_chamber_array;

	/* Start the timer. */

	mx_status = mx_timer_start( timer_record, measurement_time );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	rinfo->operation_in_progress = TRUE;

	MX_DEBUG(-2,("%s complete.", fname));

	return MX_SUCCESSFUL_RESULT;
}

/*------------------------------------------------------------------*/

static mx_status_type
stoh_register_ion_chamber( MXBL_EVENT *event )
{
	static const char fname[] = "stoh_register_ion_chamber()";

	MX_SOCKET *mx_socket;
	MX_RECORD *ion_chamber_record;
	MX_RECORD *timer_record;
	MX_ANALOG_INPUT *ainput;
	MX_SCALER *scaler;
	char *ptr, *token_ptr;
	char *mx_ion_chamber_name;
	mx_status_type mx_status;

	MX_DEBUG(-2,("%s invoked for '%s'", fname, event->message_buffer));

	mx_socket = event->mx_socket;

	/* Skip over the command name. */

	ptr = event->message_buffer;

	token_ptr = mx_string_token( &ptr, " " );

	if ( token_ptr == NULL ) {
		return mx_error( MXE_UNPARSEABLE_STRING, fname,
		"The DHS message buffer received for socket %d was empty.  "
		"This should not be able to happen.", mx_socket->socket_fd );
	}

	/* The next token should be the MX ion chamber name. */

	mx_ion_chamber_name = mx_string_token( &ptr, " " );

	if ( mx_ion_chamber_name == NULL ) {
		return mx_error( MXE_UNPARSEABLE_STRING, fname,
		"No MX ion chamber name was seen in the "
		"stoh_register_ion_chamber command received by DHS socket %d.",
			mx_socket->socket_fd );
	}

	MX_DEBUG(-2,("%s: MX ion chamber name = '%s'",
		fname, mx_ion_chamber_name ));

	/* Find the timer in the local MX database. */

	ion_chamber_record = mx_get_record( event->event_manager->record_list,
							mx_ion_chamber_name );

	if ( ion_chamber_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NOT_FOUND, fname,
		"DHS ion chamber '%s' was not found in the MX database.",
			mx_ion_chamber_name );
	}

	MX_DEBUG(-2,("%s: DHS ion chamber '%s' was _found_ in the MX database.",
		fname, ion_chamber_record->name ));

	/* Find the timer record that corresponds to this ion chamber. */

	switch( ion_chamber_record->mx_class ) {
	case MXC_ANALOG_INPUT:
		ainput = (MX_ANALOG_INPUT *)
				ion_chamber_record->record_class_struct;

		timer_record = ainput->timer_record;
		break;

	case MXC_SCALER:
		scaler = (MX_SCALER *) ion_chamber_record->record_class_struct;

		timer_record = scaler->timer_record;
		break;

	default:
		return mx_error( MXE_TYPE_MISMATCH, fname,
		"The requested MX ion chamber record '%s' is not "
		"an MX analog input record or an MX scaler record.",
			ion_chamber_record->name );
		break;
	}

	if ( timer_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_INITIALIZATION_ERROR, fname,
		"Ion chamber record '%s' does not have an associated "
		"timer record.", ion_chamber_record->name );
	}

	MX_DEBUG(-2,("%s: MX ion chamber '%s' is associated with MX timer '%s'",
		fname, ion_chamber_record->name, timer_record->name));

	/* We use the application_ptr element of the MX_RECORD structure
	 * to hold a pointer to an MXBL_RECORD_INFO object, which connects
	 * the Blu-Ice information to the MX information.
	 */

	mx_status = mxbl_setup_record_info( event->event_manager,
						timer_record,
						timer_record->name );
	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Request value changed callbacks for the "busy" field. */

	mx_status = mxbl_register_field_callback( timer_record, "busy" );

	return mx_status;
}

/*------------------------------------------------------------------*/

static mx_status_type
stoh_register_operation( MXBL_EVENT *event )
{
	static const char fname[] = "stoh_register_operation()";

	MX_SOCKET *mx_socket;
	MX_RECORD *current_record, *record_list, *operation_record;
	MX_BLUICE_SELF_OPERATION *bluice_self_operation;
	char *ptr, *token_ptr;
	char *bluice_self_operation_name;
	mx_status_type mx_status;

	MX_DEBUG(-2,("%s invoked for '%s'", fname, event->message_buffer));

	mx_socket = event->mx_socket;

	record_list = event->event_manager->record_list;

	/* Skip over the command name. */

	ptr = event->message_buffer;

	token_ptr = mx_string_token( &ptr, " " );

	if ( token_ptr == NULL ) {
		return mx_error( MXE_UNPARSEABLE_STRING, fname,
		"The DHS message buffer received for socket %d was empty.  "
		"This should not be able to happen.", mx_socket->socket_fd );
	}

	/* The next token should be the Blu-Ice operation name. */

	bluice_self_operation_name = mx_string_token( &ptr, " " );

	if ( bluice_self_operation_name == NULL ) {
		return mx_error( MXE_UNPARSEABLE_STRING, fname,
		"No Blu-Ice operation name was seen in the "
		"stoh_register_operation command received by DHS socket %d.",
			mx_socket->socket_fd );
	}

#if 0
	/* The final token should be the MX operation name. */

	mx_operation_name = mx_string_token( &ptr, " " );

	if ( mx_operation_name == NULL ) {
		return mx_error( MXE_UNPARSEABLE_STRING, fname,
		"No MX operation name was seen in the "
		"stoh_register_operation command received by DHS socket %d.",
			mx_socket->socket_fd );
	}
#endif

	MX_DEBUG(-2,("%s: Blu-Ice operation name = '%s'",
		fname, bluice_self_operation_name));

	/* Find the operation in the local MX database. */

	operation_record = NULL;

	current_record = record_list->next_record;

	while ( current_record != record_list ) {
	    if ( current_record->mx_type == MXV_BLUICE_SELF_OPERATION ) {
	    	bluice_self_operation = current_record->record_type_struct;

		MX_DEBUG(-2,
	("%s: Blu-Ice operation '%s', record_ptr = %p, bluice_name = '%s'",
			fname, current_record->name, current_record,
			bluice_self_operation->bluice_name ));

		if ( strcmp( bluice_self_operation_name,
				bluice_self_operation->bluice_name ) == 0 )
		{
		    operation_record = current_record;

		    break;	/* Exit the while() loop. */
		}
	    }

	    current_record = current_record->next_record;
	}

	if ( operation_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NOT_FOUND, fname,
		"DHS operation '%s' was not found in the MX database.",
			bluice_self_operation_name );
	}

	MX_DEBUG(-2,("%s: DHS operation '%s' was _found_ in the MX database.",
		fname, operation_record->name ));

	/* We use the application_ptr element of the MX_RECORD structure
	 * to hold a pointer to an MXBL_RECORD_INFO object, which connects
	 * the Blu-Ice information to the MX information.
	 */

	mx_status = mxbl_setup_record_info( event->event_manager,
						operation_record,
						bluice_self_operation_name );
	return mx_status;
}

/*------------------------------------------------------------------*/

static mx_status_type
stoh_register_real_motor( MXBL_EVENT *event )
{
	static const char fname[] = "stoh_register_real_motor()";

	MX_SOCKET *mx_socket;
	MX_RECORD *motor_record;
	char *ptr, *token_ptr, *bluice_motor_name, *mx_motor_name;
	char command[MXU_RECORD_NAME_LENGTH+30];
	mx_status_type mx_status;

	MX_DEBUG(-2,("%s invoked.", fname));

	mx_socket = event->mx_socket;

	/* Skip over the command name. */

	ptr = event->message_buffer;

	token_ptr = mx_string_token( &ptr, " " );

	if ( token_ptr == NULL ) {
		return mx_error( MXE_UNPARSEABLE_STRING, fname,
		"The DHS message buffer received for socket %d was empty.  "
		"This should not be able to happen.", mx_socket->socket_fd );
	}

	/* The next token should be the Blu-Ice motor name. */

	bluice_motor_name = mx_string_token( &ptr, " " );

	if ( bluice_motor_name == NULL ) {
		return mx_error( MXE_UNPARSEABLE_STRING, fname,
		"No Blu-Ice motor name was seen in the "
		"stoh_register_real_motor command received by DHS socket %d.",
			mx_socket->socket_fd );
	}

	/* The final token should be the MX motor name. */

	mx_motor_name = mx_string_token( &ptr, " " );

	if ( mx_motor_name == NULL ) {
		return mx_error( MXE_UNPARSEABLE_STRING, fname,
		"No MX motor name was seen in the "
		"stoh_register_real_motor command received by DHS socket %d.",
			mx_socket->socket_fd );
	}

	MX_DEBUG(-2,("%s: Blu-Ice motor name = '%s', MX motor name = '%s'",
		fname, bluice_motor_name, mx_motor_name ));

	/* Find the motor in the local MX database. */

	motor_record = mx_get_record( event->event_manager->record_list,
							mx_motor_name );

	if ( motor_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NOT_FOUND, fname,
		"DHS motor '%s' was not found in the MX database.",
			mx_motor_name );
	}

	MX_DEBUG(-2,("%s: DHS motor '%s' was _found_ in the MX database.",
		fname, motor_record->name ));

	/* We use the application_ptr element of the MX_RECORD structure
	 * to hold a pointer to an MXBL_RECORD_INFO object, which connects
	 * the Blu-Ice information to the MX information.
	 */

	mx_status = mxbl_setup_record_info( event->event_manager,
						motor_record,
						bluice_motor_name );
	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Request value changed callbacks for the "extended_status" field. */

	mx_status = mxbl_register_field_callback( motor_record,
						"extended_status" );
	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Request the configuration of this motor from DCSS. */

	snprintf( command, sizeof(command),
		"htos_send_configuration %s", bluice_motor_name );

	mx_status = mxbl_send_message( mx_socket, command,
					NULL, 0, MX_BLUICE_PROTOCOL_2 );

	return mx_status;
}

/*------------------------------------------------------------------*/

static mx_status_type
stoh_register_shutter( MXBL_EVENT *event )
{
	static const char fname[] = "stoh_register_shutter()";

	MX_SOCKET *mx_socket;
	MX_RECORD *shutter_record;
	char *ptr, *token_ptr;
	char *bluice_shutter_name, *mx_shutter_name;
	mx_status_type mx_status;

	MX_DEBUG(-2,("%s invoked for '%s'", fname, event->message_buffer));

	mx_socket = event->mx_socket;

	/* Skip over the command name. */

	ptr = event->message_buffer;

	token_ptr = mx_string_token( &ptr, " " );

	if ( token_ptr == NULL ) {
		return mx_error( MXE_UNPARSEABLE_STRING, fname,
		"The DHS message buffer received for socket %d was empty.  "
		"This should not be able to happen.", mx_socket->socket_fd );
	}

	/* The next token should be the Blu-Ice shutter name. */

	bluice_shutter_name = mx_string_token( &ptr, " " );

	if ( bluice_shutter_name == NULL ) {
		return mx_error( MXE_UNPARSEABLE_STRING, fname,
		"No Blu-Ice shutter name was seen in the "
		"stoh_register_shutter command received by DHS socket %d.",
			mx_socket->socket_fd );
	}

	/* The final token should be the MX shutter name. */

	mx_shutter_name = mx_string_token( &ptr, " " );

	if ( mx_shutter_name == NULL ) {
		return mx_error( MXE_UNPARSEABLE_STRING, fname,
		"No MX shutter name was seen in the "
		"stoh_register_shutter command received by DHS socket %d.",
			mx_socket->socket_fd );
	}

	MX_DEBUG(-2,("%s: Blu-Ice shutter name = '%s', MX shutter name = '%s'",
		fname, bluice_shutter_name, mx_shutter_name ));

	/* Find the shutter in the local MX database. */

	shutter_record = mx_get_record( event->event_manager->record_list,
							mx_shutter_name );

	if ( shutter_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NOT_FOUND, fname,
		"DHS shutter '%s' was not found in the MX database.",
			mx_shutter_name );
	}

	MX_DEBUG(-2,("%s: DHS shutter '%s' was _found_ in the MX database.",
		fname, shutter_record->name ));

	/* We use the application_ptr element of the MX_RECORD structure
	 * to hold a pointer to an MXBL_RECORD_INFO object, which connects
	 * the Blu-Ice information to the MX information.
	 */

	mx_status = mxbl_setup_record_info( event->event_manager,
						shutter_record,
						bluice_shutter_name );
	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Request value changed callbacks for the "relay_status" field. */

	mx_status = mxbl_register_field_callback( shutter_record,
						"relay_status" );

	return mx_status;
}

/*------------------------------------------------------------------*/

static mx_status_type
stoh_register_string( MXBL_EVENT *event )
{
	static const char fname[] = "stoh_register_string()";

	MX_SOCKET *mx_socket;
	MX_RECORD *current_record, *record_list, *string_record;
	MX_BLUICE_STRING *bluice_string;
	char *ptr, *token_ptr;
	char *bluice_string_name;
	mx_status_type mx_status;

	MX_DEBUG(-2,("%s invoked for '%s'", fname, event->message_buffer));

	mx_socket = event->mx_socket;

	record_list = event->event_manager->record_list;

	/* Skip over the command name. */

	ptr = event->message_buffer;

	token_ptr = mx_string_token( &ptr, " " );

	if ( token_ptr == NULL ) {
		return mx_error( MXE_UNPARSEABLE_STRING, fname,
		"The DHS message buffer received for socket %d was empty.  "
		"This should not be able to happen.", mx_socket->socket_fd );
	}

	/* The next token should be the Blu-Ice string name. */

	bluice_string_name = mx_string_token( &ptr, " " );

	if ( bluice_string_name == NULL ) {
		return mx_error( MXE_UNPARSEABLE_STRING, fname,
		"No Blu-Ice string name was seen in the "
		"stoh_register_string command received by DHS socket %d.",
			mx_socket->socket_fd );
	}

	/* FIXME: If the maximum length of MX record names was longer, then
	 * Blu-Ice strings could be ordinary strings in this context, rather
	 * than bluice_self_string records.
	 */
#if 0
	/* The final token should be the MX string name. */

	mx_string_name = mx_string_token( &ptr, " " );

	if ( mx_string_name == NULL ) {
		return mx_error( MXE_UNPARSEABLE_STRING, fname,
		"No MX string name was seen in the "
		"stoh_register_string command received by DHS socket %d.",
			mx_socket->socket_fd );
	}
#endif

	MX_DEBUG(-2,("%s: Blu-Ice string name = '%s'",
		fname, bluice_string_name));

	/* Find the string in the local MX database. */

	string_record = NULL;

	current_record = record_list->next_record;

	while ( current_record != record_list ) {
	    if ( current_record->mx_type == MXV_BLUICE_SELF_STRING ) {
	    	bluice_string = current_record->record_type_struct;

		if ( strcmp( bluice_string_name,
				bluice_string->bluice_name ) == 0 )
		{
		    string_record = current_record;

		    break;	/* Exit the while() loop. */
		}
	    }

	    current_record = current_record->next_record;
	}

	if ( string_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NOT_FOUND, fname,
		"DHS string '%s' was not found in the MX database.",
			bluice_string_name );
	}

	MX_DEBUG(-2,("%s: DHS string '%s' was _found_ in the MX database.",
		fname, string_record->name ));

	/* We use the application_ptr element of the MX_RECORD structure
	 * to hold a pointer to an MXBL_RECORD_INFO object, which connects
	 * the Blu-Ice information to the MX information.
	 */

	mx_status = mxbl_setup_record_info( event->event_manager,
						string_record,
						bluice_string_name );
	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Request value changed callbacks for the "value" field. */

	mx_status = mxbl_register_field_callback( string_record, "value" );

	return mx_status;
}

/*------------------------------------------------------------------*/

static mx_status_type
stoh_set_motor_position( MXBL_EVENT *event )
{
	static const char fname[] = "stoh_set_motor_position()";

	MX_SOCKET *mx_socket;
	MX_RECORD *record_list, *motor_record;
	MXBL_RECORD_INFO *rinfo;
	char *endptr, *ptr, *token_ptr, *motor_name, *set_position_string;
	double set_position;
	mx_status_type mx_status;

	MX_DEBUG(-2,("%s invoked.", fname));

	mx_socket = event->mx_socket;

	record_list = event->event_manager->record_list;

	/* Skip over the command name. */

	ptr = event->message_buffer;

	token_ptr = mx_string_token( &ptr, " " );

	if ( token_ptr == NULL ) {
		return mx_error( MXE_UNPARSEABLE_STRING, fname,
		"The DHS message buffer received for socket %d was empty.  "
		"This should not be able to happen.", mx_socket->socket_fd );
	}

	/* The next token should be the motor name. */

	motor_name = mx_string_token( &ptr, " " );

	if ( motor_name == NULL ) {
		return mx_error( MXE_UNPARSEABLE_STRING, fname,
		"No motor name was seen in the stoh_set_motor_position command "
		"received by DHS socket %d.", mx_socket->socket_fd );
	}

	/* The last token should be the new motor position. */

	set_position_string = mx_string_token( &ptr, " " );

	if ( set_position_string == NULL ) {
		return mx_error( MXE_UNPARSEABLE_STRING, fname,
		"No new position was specified for motor '%s' by the "
		"stoh_set_motor_position command received by DHS socket %d.",
			motor_name, mx_socket->socket_fd );
	}

	/* Find the motor in the MX database. */

	motor_record = mx_get_record( record_list, motor_name );

	if ( motor_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NOT_FOUND, fname,
		"DHS socket %d attempted to redefine the position of "
		"motor '%s', but there is no motor by that name.",
			mx_socket->socket_fd, motor_name );
	}

	set_position = strtod( set_position_string, &endptr );

	if ( *endptr != '\0' ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"The new position '%s' requested by DHS socket %d "
		"for motor '%s' is not a legal floating point number.",
			set_position_string, mx_socket->socket_fd,
			motor_record->name );
	}

	rinfo = motor_record->application_ptr;

	if ( rinfo == (MXBL_RECORD_INFO *) NULL ) {
		return mx_error( MXE_INITIALIZATION_ERROR, fname,
		"No MXBL_RECORD_INFO structure has been allocated "
		"for record '%s'.", motor_record->name );
	}

	/* Redefine the current position. */

	MX_DEBUG(-2,("%s: Redefining the position of '%s' as %g",
		fname, motor_record->name, set_position));

	mx_status = mx_motor_set_position( motor_record, set_position );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	rinfo->operation_in_progress = TRUE;

	/* We do not send a message to DCSS here.  Instead, we rely on
	 * the MX position callback to see that operation_in_progress
	 * is set and then update the status of DCSS.
	 */

	return MX_SUCCESSFUL_RESULT;
}

/*------------------------------------------------------------------*/

void
mxbl_handle_dhs_messages( MXBL_EVENT *event )
{
	static const char fname[] = "mxbl_handle_dhs_messages()";

	MX_SOCKET *mx_socket;
	MXBL_DHS_MSG_HANDLER *message_fn;
	double receive_timeout_in_seconds;
	double loop_timeout_in_seconds;
	MX_CLOCK_TICK loop_timeout_in_ticks;
	MX_CLOCK_TICK finish_tick, current_tick;
	int i, num_items, tick_comparison;
	long actual_data_length;
	char message_type_name[80];
	char message_type_format[20];
	mx_bool_type socket_is_open;
	mx_status_type mx_status;

	if ( event->protocol_initialized == FALSE ) {
		mxbl_handle_dhs_message_initialization( event );
	}

	mx_socket = event->mx_socket;

	receive_timeout_in_seconds = 0.01;

	loop_timeout_in_seconds = 0.1;

	loop_timeout_in_ticks = mx_convert_seconds_to_clock_ticks(
						loop_timeout_in_seconds );

	snprintf( message_type_format, sizeof(message_type_format),
			"%%%lus", (unsigned long) sizeof(message_type_name) );

	MX_DEBUG(-2,("%s: DHS socket %d is initialized.",
		fname, mx_socket->socket_fd ));

	current_tick = mx_current_clock_tick();

	finish_tick = mx_add_clock_ticks( current_tick, loop_timeout_in_ticks );

	for (;;) {
		MX_DEBUG(-2,("%s: Top of the for(;;) loop.", fname));

		/* See if we have timed out. */

		current_tick = mx_current_clock_tick();

		tick_comparison = mx_compare_clock_ticks( current_tick,
							finish_tick );

		if ( tick_comparison >= 0 ) {
			MX_DEBUG(-2,
			("%s: Exiting for(;;) loop after %g seconds.",
				fname, loop_timeout_in_seconds));

			break;	/* Exit the for(;;) loop. */
		}

		/* Give up if the socket is not open. */

		socket_is_open = mx_socket_is_open( mx_socket );

		if ( socket_is_open == FALSE ) {
			MX_DEBUG(-2,("%s: The DHS socket is not open.", fname));

			mx_msleep(100);

			/* Go back to the top of the for(;;) loop. */

			continue;
		}

		/* Try to read a message from DCSS. */

		mx_status = mxbl_receive_message( mx_socket,
					event->message_buffer,
					sizeof(event->message_buffer),
					&actual_data_length,
					receive_timeout_in_seconds,
					MX_BLUICE_PROTOCOL_2 );

		if ( mx_status.code == MXE_NETWORK_CONNECTION_LOST ) {

			(void) mxbl_event_manager_unregister_socket_event(
								event );

			mx_socket_close( mx_socket );

			return;
		} else
		if ( mx_status.code != MXE_SUCCESS ) {
			mx_msleep(100);

			continue;
		}

#if 0
		MX_DEBUG(-2,("%s: received '%s' from DHS socket %d.",
			fname, event->message_buffer, mx_socket->socket_fd ));
#endif

		/* Figure out what kind of message this is. */

		num_items = sscanf( event->message_buffer,
				message_type_format,
				message_type_name );

		for ( i = 0; i < num_dhs_handlers; i++ ) {
			if ( strcmp( message_type_name,
				dhs_handler_list[i].message_name ) == 0 )
			{
				break;		/* Exit the for(i) loop. */
			}
		}

		if ( i >= num_dhs_handlers ) {
			(void) mx_error( MXE_NETWORK_IO_ERROR, fname,
			"Message type '%s' received in message '%s' from "
			"DHS socket %d does not currently have a handler.",
				message_type_name, event->message_buffer,
				mx_socket->socket_fd );
		} else {
			message_fn = dhs_handler_list[i].message_handler;

			if ( message_fn == NULL ) {
				MX_DEBUG(-2,("%s: Message type '%s' SKIPPED.",
					fname, message_type_name));
			} else {
				MX_DEBUG(-2,("%s: Invoking handler for '%s'",
					fname, message_type_name));

				mx_status = (*message_fn)( event );
			}
		}
	}

	return;
}

