/*
 * Name:    mxbl_event.c
 *
 * Purpose: Event management for the MX Blu-Ice server.
 *
 * Author:  William Lavender
 *
 *--------------------------------------------------------------------------
 *
 * Copyright 2008 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>

#include "mx_osdef.h"
#include "mx_util.h"
#include "mx_record.h"
#include "mx_stdint.h"
#include "mx_pipe.h"
#include "mx_socket.h"
#include "mx_select.h"
#include "mx_list.h"
#include "mx_callback.h"

#include "mx_poison.h"

#include "mxbl.h"

/*------------------------------------------------------------------*/

mx_status_type
mxbl_create_event_manager( MXBL_EVENT_MANAGER **event_manager,
			MX_RECORD *record_list,
			char *dhs_name )
{
	static const char fname[] = "mxbl_create_event_manager()";

	mx_status_type mx_status;

	if ( event_manager == (MXBL_EVENT_MANAGER **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MXBL_EVENT_MANAGER pointer passed was NULL." );
	}

	*event_manager = (MXBL_EVENT_MANAGER *)
				calloc( 1, sizeof(MXBL_EVENT_MANAGER) );

	if ( (*event_manager) == (MXBL_EVENT_MANAGER *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Ran out of memory trying to allocate "
		"an MXBL_EVENT_MANAGER structure." );
	}

	(*event_manager)->record_list = record_list;

	strlcpy( (*event_manager)->dhs_name, dhs_name,
		sizeof( (*event_manager)->dhs_name ) );

	mx_status = mx_list_create( &((*event_manager)->event_list) );

	return mx_status;
}

mx_status_type
mxbl_event_manager_register_socket_event(
				MXBL_EVENT_MANAGER *event_manager,
				MX_SOCKET *mx_socket,
				long socket_type,
				void (*event_handler)(MXBL_EVENT *),
				MXBL_EVENT **new_event )
{
	static const char fname[] =
		"mxbl_event_manager_register_socket_event()";

	MXBL_EVENT *event;
	MX_LIST_ENTRY *new_list_entry;
	mx_status_type mx_status;

	if ( event_manager == (MXBL_EVENT_MANAGER *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MXBL_EVENT_MANAGER pointer passed was NULL." );
	}
	if ( mx_socket == (MX_SOCKET *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_SOCKET pointer passed was NULL." );
	}

	MX_DEBUG(-2,("%s: mx_socket = %p", fname, mx_socket));
	MX_DEBUG(-2,("%s: mx_socket->socket_fd = %d",
				fname, mx_socket->socket_fd));

	event = (MXBL_EVENT *) malloc( sizeof(MXBL_EVENT) );

	if ( event == (MXBL_EVENT *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Ran out of memory trying to allocate "
		"an MXBL_EVENT structure." );
	}

	MX_DEBUG(-2,("%s: event = %p", fname, event));

	event->socket_type = socket_type;
	event->mx_socket = mx_socket;
	event->event_handler = event_handler;
	event->event_manager = event_manager;
	event->protocol_initialized = FALSE;

	mx_status = mx_list_entry_create( &new_list_entry, event, NULL );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mx_status = mx_list_add_entry( event_manager->event_list,
						new_list_entry );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

#if 0
	mx_info(
    "%s: event_manager = %p, event_list = %p, new_list_entry = %p, event = %p",
	    	fname, event_manager, event_manager->event_list,
		new_list_entry, event );

	mxbl_event_manager_list_socket_events( event_manager,
	"After adding new event:" );
#endif

	if ( new_event != (MXBL_EVENT **) NULL ) {
		*new_event = event;
	}

	return MX_SUCCESSFUL_RESULT;
}

mx_status_type
mxbl_event_manager_unregister_socket_event( MXBL_EVENT *event )
{
	static const char fname[] =
		"mxbl_event_manager_unregister_socket_event()";

	MXBL_EVENT_MANAGER *event_manager;
	MX_SOCKET *mx_socket;
	MX_LIST_ENTRY *list_entry;
	mx_status_type mx_status;

	if ( event == (MXBL_EVENT *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MXBL_EVENT pointer passed was NULL." );
	}

	MX_DEBUG(-2,("%s invoked for event %p", fname, event));

	event_manager = event->event_manager;

	mx_socket = event->mx_socket;

	MX_DEBUG(-2,("%s: socket = %d", fname, mx_socket->socket_fd));

	/* Was this event a Blu-Ice master? */

	if ( ( event->socket_type == MXBLT_DCSS_CLIENT_SOCKET )
	  && ( event == event_manager->master ) )
	{
		/* If so, then record the fact that we no longer have
		 * a Blu-Ice master client.
		 */

		event_manager->master = NULL;

		MX_DEBUG(-2,("%s: DCSS no longer has a master.", fname));
	}

	/* Delete the event from the event list. */

	mx_status = mx_list_find_list_entry( event_manager->event_list,
					event, &list_entry );

	if ( mx_status.code == MXE_NOT_FOUND ) {
		return mx_error( MXE_NOT_FOUND, fname,
		"MX_SOCKET %p, socket %d was not found in the event list.",
			mx_socket, mx_socket->socket_fd );
	} else
	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

#if 0
	mx_info("%s: event_manager = %p, event_list = %p, "
	"list_entry_to_delete = %p, event = %p",
		fname, event_manager, event_manager->event_list,
		list_entry, event );
#endif

	mx_status = mx_list_delete_entry( event_manager->event_list,
							list_entry );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	event = list_entry->list_entry_data;

	mx_free(list_entry);
	mx_free(event);

	return MX_SUCCESSFUL_RESULT;
}

static mx_status_type
mxblp_add_socket( MX_LIST_ENTRY *list_entry, void *args, void **not_used )
{
	static const char fname[] = "mxblp_add_socket()";

	MXBL_EVENT *event;
	MX_SOCKET *mx_socket;
	MXBL_EVENT_MANAGER *event_manager;
	fd_set *readfds;

	MX_DEBUG( 2,("%s: list_entry = %p, args = %p",
		fname, list_entry, args));

	readfds = args;

	event = list_entry->list_entry_data;

	MX_DEBUG( 2,("%s: event = %p", fname, event));

	mx_socket = event->mx_socket;

	MX_DEBUG( 2,("%s: mx_socket = %p", fname, mx_socket));
	MX_DEBUG( 2,("%s: mx_socket->socket_fd = %d",
			fname, mx_socket->socket_fd));

	FD_SET( mx_socket->socket_fd, readfds );

	MX_DEBUG( 2,("%s: list_entry = %p, socket = %d",
		fname, list_entry, mx_socket->socket_fd));

	event_manager = event->event_manager;

	if ( mx_socket->socket_fd > event_manager->max_fd ) {
		event_manager->max_fd = mx_socket->socket_fd;
	}

	if ( event->socket_type == MXBLT_DHS_SOCKET ) {
		if ( event_manager->dhs_socket_is_present ) {
			(void) mx_error( MXE_UNKNOWN_ERROR, fname,
			"While processing socket %d, it was found that "
			"there are more than one socket in the event list "
			"that are of type MXBLT_DHS_SOCKET.  "
			"This should never happen.", mx_socket->socket_fd );
		}

		event_manager->dhs_socket_is_present = TRUE;
	}

	return MX_SUCCESSFUL_RESULT;
}

static mx_status_type
mxblp_process_socket( MX_LIST_ENTRY *list_entry, void *args, void **not_used )
{
	static const char fname[] = "mxblp_process_socket()";

	MXBL_EVENT *event;
	MX_SOCKET *mx_socket;
	fd_set *readfds;
	int isset;
	void (*event_handler)(MXBL_EVENT *);

	MX_DEBUG( 2,("%s: list_entry = %p, args = %p",
		fname, list_entry, args));

	readfds = args;

	event = list_entry->list_entry_data;

	MX_DEBUG( 2,("%s: event = %p", fname, event));

	mx_socket = event->mx_socket;

	MX_DEBUG( 2,("%s: mx_socket = %p", fname, mx_socket));
	MX_DEBUG( 2,("%s: mx_socket->socket_fd = %d",
			fname, mx_socket->socket_fd));

	isset = FD_ISSET( mx_socket->socket_fd, readfds );

	MX_DEBUG( 2,("%s: isset = %d", fname, isset));

	if ( isset == FALSE ) {
		/* This socket had no activity, so return now. */

		return MX_SUCCESSFUL_RESULT;
	}

	/* This socket _did_ have activity, so invoke the event handler. */

	event_handler = event->event_handler;

	MX_DEBUG( 2,("%s: event_handler = %p", fname, event_handler));

	if ( event_handler != NULL ) {
		(*event_handler)( event );
	}

	return MX_SUCCESSFUL_RESULT;
}

mx_status_type
mxbl_event_manager_process_events( MXBL_EVENT_MANAGER *event_manager,
					double timeout_in_seconds )
{
	static const char fname[] = "mxbl_event_manager_process_events()";

	int num_fds, max_fd, saved_errno;
	fd_set readfds;
	struct timeval timeout;
	mx_status_type mx_status;

	if ( event_manager == (MXBL_EVENT_MANAGER *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MXBL_EVENT_MANAGER pointer passed was NULL." );
	}

	/* Initialize the arguments to select(). */

	FD_ZERO(&readfds);

	event_manager->max_fd = -1;
	event_manager->dhs_socket_is_present = FALSE;

	mx_status = mx_list_traverse( event_manager->event_list,
					mxblp_add_socket, &readfds, NULL );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

#if defined(OS_WIN32)
	max_fd = -1;
#else
	max_fd = event_manager->max_fd + 1;
#endif

	timeout.tv_sec = timeout_in_seconds;
	timeout.tv_usec = 1000000.0 * ( timeout_in_seconds - timeout.tv_sec );

	MX_DEBUG( 2,("%s: calling select(%d,...)", fname, max_fd));

	num_fds = select( max_fd, &readfds, NULL, NULL, &timeout );

	saved_errno = errno;

	MX_DEBUG( 2,("%s: num_fds = %d", fname, num_fds));

	if ( num_fds < 0 ) {
		if ( saved_errno == EINTR ) {
#if 1
			MX_DEBUG(-2,("%s: EINTR returned by select()",
				fname));
#endif
			/* Receiving an EINTR errno from select() is normal.
			 * It just means that a signal handler fired while
			 * we were blocked in the select() system call.
			 */
		} else {
			(void) mx_error( MXE_NETWORK_IO_ERROR, fname,
			"Error in select() while waiting for events.  "
			"Errno = %d.  Error string = '%s'",
			saved_errno, strerror(saved_errno) );
		}
	} else if ( num_fds == 0 ) {

		/* Did not get any events, so do nothing here. */
	} else {
		/* Figure out which sockets had events and then
		 * process the events.
		 */

#if 0
		mxbl_event_manager_list_socket_events( event_manager,
		"Before list traverse in mxbl_event_manager_process_events():");
#endif
		
		mx_status = mx_list_traverse( event_manager->event_list,
					mxblp_process_socket, &readfds, NULL );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;
	}

	return MX_SUCCESSFUL_RESULT;
}

mx_status_type
mxbl_event_manager_list_socket_events( MXBL_EVENT_MANAGER *event_manager,
					char *label )
{
	static const char fname[] = "mxbl_event_manager_list_socket_events()";

	MX_LIST *event_list;
	MX_LIST_ENTRY *list_start, *list_entry;
	MXBL_EVENT *event;

	if ( event_manager == (MXBL_EVENT_MANAGER *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MXBL_EVENT_MANAGER pointer passed was NULL." );
	}

	if ( label != NULL ) {
		mx_info( label );
	} else {
		mx_info( "Socket event list:" );
	}

	event_list = event_manager->event_list;

	mx_info( "  event_manager = %p, event_list = %p",
		event_manager, event_list );

	if ( event_list == NULL ) {
		mx_info( "  Event list is NULL." );
		return MX_SUCCESSFUL_RESULT;
	}

	list_start = event_list->list_start;

	if ( event_list == NULL ) {
		mx_info( "  Event list is empty." );
	}

	list_entry = list_start;

	do {
		event = list_entry->list_entry_data;

		mx_info( "  list_entry = %p, event = %p", list_entry, event );

		list_entry = list_entry->next_list_entry;

	} while ( list_entry != list_start );

	return MX_SUCCESSFUL_RESULT;
}

