/*
 * Name:    mxbl_main.c
 *
 * Purpose: main() routine for the MX Blu-Ice server.
 *
 * Author:  William Lavender
 *
 *--------------------------------------------------------------------------
 *
 * Copyright 2008 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>

#include "mx_osdef.h"
#include "mx_util.h"
#include "mx_record.h"
#include "mx_version.h"
#include "mx_unistd.h"
#include "mx_stdint.h"
#include "mx_cfn.h"
#include "mx_pipe.h"
#include "mx_socket.h"
#include "mx_callback.h"
#include "mx_process.h"

#include "mx_poison.h"

#include "mxbl.h"

/*------------------------------------------------------------------*/

static int
mxbl_user_interrupt_function( void )
{
	/* The MX Blu-Ice server never has direct user interrupts. */

	return MXF_USER_INT_NONE;
}

static void
mxbl_print_timestamp( void )
{
	time_t time_struct;
	struct tm *current_time;
	char buffer[80];

	time( &time_struct );

	current_time = localtime( &time_struct );

	strftime( buffer, sizeof(buffer),
			"%b %d %H:%M:%S ", current_time );

	fputs( buffer, stderr );
}

static void
mxbl_output_function( char *string )
{
	mxbl_print_timestamp();

	fprintf( stderr, "%s\n", string );
}

static void
mxbl_setup_output_functions( void )
{
	mx_set_info_output_function( mxbl_output_function );
	mx_set_warning_output_function( mxbl_output_function );
	mx_set_error_output_function( mxbl_output_function );
	mx_set_debug_output_function( mxbl_output_function );
}

/*------------------------------------------------------------------*/

int
main( int argc, char *argv[] )
{
	static const char fname[] = "main()";

	MX_RECORD *record_list;
	MX_LIST_HEAD *list_head;
	MX_PIPE *callback_pipe;
	MX_SOCKET *dcss_socket, *dhs_socket;
	MXBL_EVENT_MANAGER *event_manager;
	MXBL_EVENT *dcss_server_event;
	char dhs_address[80];
	char dhs_name[MXU_BLUICE_NAME_LENGTH+1];
	char *dhs_hostname, *ptr;
	int dcss_port, dhs_port;
	int debug_level;
	size_t num_bytes_available;
	mx_bool_type start_debugger;
	char mx_database_filename[MXU_FILENAME_LENGTH+1];
	mx_status_type mx_status;

#if HAVE_GETOPT
	int c, error_flag;
#endif

	/* Initialize the parts of the MX runtime environment that do not
	 * depend on the presence of an MX database.
	 */

	mx_status = mx_initialize_runtime();

	if ( mx_status.code != MXE_SUCCESS ) {
		fprintf( stderr,
		"%s: Unable to initialize MX runtime environment.\n",
			argv[0] );

		exit(1);
	}

	mxbl_setup_output_functions();

	mx_construct_config_filename( "mxbluice.dat",
			mx_database_filename, MXU_FILENAME_LENGTH );

	/* Set default values for command line switches. */

	debug_level = 0;
	start_debugger = FALSE;

	dcss_port = -1;

	strlcpy( dhs_address, "", sizeof(dhs_address) );

	strlcpy( dhs_name, "mx", sizeof(dhs_name) );

#if HAVE_GETOPT
	/* Process command line arguments, if any. */

	error_flag = FALSE;

	while ((c = getopt(argc,argv,"c:d:Df:h:n:")) != -1)
	{
		switch (c) {
		case 'c':
			dcss_port = atoi( optarg );
			break;
		case 'd':
			debug_level = atoi( optarg );
			break;
		case 'D':
			start_debugger = TRUE;
			break;
		case 'f':
			strlcpy( mx_database_filename, optarg,
					sizeof(mx_database_filename) );
			break;
		case 'h':
			strlcpy( dhs_address, optarg, sizeof(dhs_address) );
			break;
		case 'n':
			strlcpy( dhs_name, optarg, sizeof(dhs_name) );
			break;
		case '?':
			error_flag = TRUE;
			break;
		}

		if ( error_flag == TRUE ) {
			fprintf( stderr,
"Usage: mx_bluice_server [-c dcss_port] [-d debug_level] [-D] \n"
"                          [-f mx_database_filename] [-h dhs_address]\n" );
    			exit(1);
		}
	}

	if ( start_debugger ) {
		mx_start_debugger( NULL );
	}

#endif /* HAVE_GETOPT */

	mx_set_user_interrupt_function( mxbl_user_interrupt_function );

	if ( strlen(dhs_address) == 0 ) {
		dhs_hostname = NULL;
		dhs_port = -1;
	} else {
		ptr = dhs_address;

		dhs_hostname = mx_string_token( &ptr, "@" );

		if ( dhs_hostname == NULL ) {
			(void) mx_error( MXE_UNEXPECTED_END_OF_DATA, fname,
			"No DHS hostname was specified.  Exiting..." );

			exit(1);
		}

		if ( ptr == NULL ) {
			dhs_port = 14242;
		} else {
			dhs_port = atoi( ptr );
		}
	}

	if ( (dcss_port < 0) && (dhs_port < 0) ) {
		(void) mx_error( MXE_INITIALIZATION_ERROR, fname,
		"No DCSS port number or DHS host address were supplied, "
		"so the Blu-Ice server cannot start." );

		exit(1);
	}

	/*** Start setting up the MX database. ***/

	mx_info( "***** MX Blu-Ice server %s started *****",
		mx_get_version_string() );

	mx_status = mx_create_empty_database( &record_list );

	if ( mx_status.code != MXE_SUCCESS )
		exit(mx_status.code);

	mx_set_debug_level( debug_level );

	list_head = mx_get_record_list_head_struct( record_list );

	if ( list_head == (MX_LIST_HEAD *) NULL ) {
		mx_status = mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"The MX_LIST_HEAD structure for the record "
			"list is NULL." );

		exit(mx_status.code );
	}

	/*** Initialize the callback pipe. ***/

	mx_status = mx_setup_callback_pipe( record_list, &callback_pipe );

	if ( mx_status.code != MXE_SUCCESS )
		exit(mx_status.code);

	/*** Load the MX database and then finish the database setup. ***/

	mx_info("Loading MX database '%s'", mx_database_filename);
	
	mx_status = mx_read_database_file( record_list,
					mx_database_filename, 0 );

	if ( mx_status.code != MXE_SUCCESS )
		exit( mx_status.code );

	mx_status = mx_finish_database_initialization( record_list );

	if ( mx_status.code != MXE_SUCCESS )
		exit( mx_status.code );

	mx_status = mx_initialize_hardware( record_list, 0 );

	if ( mx_status.code != MXE_SUCCESS )
		exit( mx_status.code );

	/* Enable processing callbacks. */

	mx_status = mx_initialize_database_processing( record_list );

	if ( mx_status.code != MXE_SUCCESS )
		exit( mx_status.code );

	/*** The MX database is now ready to be used. ***/

	if ( dhs_port < 0 ) {
		mx_info( "DCSS port = %d", dcss_port );
	} else
	if ( dcss_port < 0 ) {
		mx_info( "DHS port = %s@%d", dhs_hostname, dhs_port );
	} else {
		mx_info( "DCSS port = %d, DHS port = %s@%d",
			dcss_port, dhs_hostname, dhs_port );
	}

	/*** Set up an event manager structure. ***/

	mx_status = mxbl_create_event_manager( &event_manager,
						record_list, dhs_name );

	if ( mx_status.code != MXE_SUCCESS )
		exit( mx_status.code );

	/*** If necessary, open the DCSS server socket. ***/

	if ( dcss_port < 0 ) {
		dcss_socket = NULL;
	} else {
		mx_status = mxbl_register_dcss_callbacks( event_manager,
							record_list );

		if ( mx_status.code != MXE_SUCCESS )
			exit( mx_status.code );

		mx_status = mx_tcp_socket_open_as_server(
					&dcss_socket, dcss_port, 
					MXF_SOCKET_DISABLE_NAGLE_ALGORITHM,
					MX_SOCKET_DEFAULT_BUFFER_SIZE );

		if ( mx_status.code != MXE_SUCCESS )
			exit( mx_status.code );

		mx_status = mxbl_event_manager_register_socket_event(
					event_manager, dcss_socket,
					MXBLT_DCSS_SERVER_SOCKET,
					mxbl_handle_dcss_server_socket,
					&dcss_server_event );

		if ( mx_status.code != MXE_SUCCESS )
			exit( mx_status.code );

		dcss_server_event->client_number = 0;
	}

	/*** If necessary, open the DHS client socket. ***/

	if ( dhs_port >= 0 ) {
		mx_status = mx_tcp_socket_open_as_client( &dhs_socket,
					dhs_hostname, dhs_port,
					MXF_SOCKET_DISABLE_NAGLE_ALGORITHM,
					MX_SOCKET_DEFAULT_BUFFER_SIZE );

		/* It is not an error if the DHS client socket fails to
		 * connect.  If this happens, it may mean that the DCSS
		 * process it is trying to connect to is not running,
		 * so we will have to try again later.
		 */

		if ( mx_status.code == MXE_SUCCESS ) {
			mx_status = mxbl_event_manager_register_socket_event(
					event_manager, dhs_socket,
					MXBLT_DHS_SOCKET,
					mxbl_handle_dhs_messages,
					NULL );

			if ( mx_status.code != MXE_SUCCESS )
				exit( mx_status.code );
		}
	}

	/*** Loop forever processing events. */

	mx_info( "Ready to accept client connections." );

	while(1) {
		mx_status = mxbl_event_manager_process_events(
					event_manager, 0.001 );

		if ( (dhs_port >= 0)
		  && ( event_manager->dhs_socket_is_present == FALSE ) )
		{
			mx_status = mx_tcp_socket_open_as_client( &dhs_socket,
					dhs_hostname, dhs_port,
					MXF_SOCKET_DISABLE_NAGLE_ALGORITHM,
					MX_SOCKET_DEFAULT_BUFFER_SIZE );

			/* It is not an error if the DHS client socket fails
			 * to connect.  If this happens, it may mean that the
			 * DCSS process it is trying to connect to is not
			 * running, so we will have to try again later.
			 */

			if ( mx_status.code == MXE_SUCCESS ) {
				mx_status =
				    mxbl_event_manager_register_socket_event(
						event_manager, dhs_socket,
						MXBLT_DHS_SOCKET,
						mxbl_handle_dhs_messages,
						NULL );

				if ( mx_status.code != MXE_SUCCESS )
					exit( mx_status.code );
			}
		}

		/* See if any callbacks have been requested. */

		if ( list_head->callback_pipe != NULL ) {

			mx_status = mx_pipe_num_bytes_available(
						list_head->callback_pipe,
						&num_bytes_available );

			if ( ( mx_status.code == MXE_SUCCESS )
			  && ( num_bytes_available > 0 ) )
			{
				mx_status = mx_process_callbacks( record_list,
						list_head->callback_pipe );
			}
		}
	}

#if !defined(OS_SOLARIS)
	exit(0);
#endif

}

